import dataclasses
import sys
from contextlib import suppress
from enum import Enum
from functools import wraps
from typing import Any, Optional, Type, TypeVar, Union

# pylint: disable=unused-import
from .arg import (
    convert_to_command_string,
    convert_to_flag_string,
    make_dataclass_from_func,
)
from .exceptions import raise_unsupported_type_error
from .types import Dataclass

__all__ = [
    "convert_to_command_string",
    "convert_to_flag_string",
    "make_dataclass_from_func",
    "create_pydantic_model_from_dataclass",
    "is_dataclass_type",
    "is_pydantic_available",
    "is_shtab_available",
    "cast_bool",
    "cast_type",
]

try:
    from shtab import add_argument_to
except ModuleNotFoundError:

    def add_argument_to():
        ...


try:
    from pydantic import parse_obj_as
    from pydantic.dataclasses import create_pydantic_model_from_dataclass
except ModuleNotFoundError:

    def parse_obj_as():
        ...

    def create_pydantic_model_from_dataclass():
        ...


if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard

T = TypeVar("T", bound=type)


def is_dataclass_type(candidate: Any) -> TypeGuard[Type[Dataclass]]:
    return dataclasses.is_dataclass(candidate)


@wraps(isinstance)
def try_isinstance(*args: Any, **kwargs: Any) -> bool:
    with suppress(TypeError):
        return isinstance(*args, **kwargs)
    return False


@wraps(issubclass)
def try_issubclass(*args: Any, **kwargs: Any) -> bool:
    with suppress(TypeError):
        return issubclass(*args, **kwargs)
    return False


def coalesce(x: Any, d: Any, null_or_empty: bool = False) -> Any:
    if (null_or_empty and x) or (not null_or_empty and x is not None):
        return x
    return d


def is_pydantic_available() -> bool:
    return create_pydantic_model_from_dataclass.__module__ != __name__


def is_shtab_available() -> bool:
    return add_argument_to.__module__ != __name__


def cast_bool(value: Union[None, str, bool]) -> bool:
    if isinstance(value, bool):
        return value

    if not isinstance(value, str):
        return bool(value)

    value_lower: str = value.lower()

    if value_lower in ("1", "true", "t", "yes", "y", "on"):
        return True

    if value_lower in ("0", "false", "f", "no", "n", "off"):
        return False

    raise ValueError(f"Invalid literal for bool(): {value}")


def cast_type(value: Any, target_type: Optional[T]) -> Optional[T]:
    if value is None or target_type is None or try_isinstance(value, target_type):
        return value

    if target_type is bool:
        return cast_bool(value)

    if try_issubclass(target_type, Enum):
        return target_type[value]

    try:
        return target_type(value)
    except TypeError as e:
        if is_pydantic_available():
            return parse_obj_as(target_type, value)

        raise_unsupported_type_error(type(value), from_exception=e)



import argparse
import collections.abc
import sys
from collections import defaultdict
from dataclasses import MISSING, Field, fields
from enum import Enum
from functools import partial, wraps
from inspect import signature
from types import GeneratorType
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .actions import (
    _split_csv_to_dict,
    split_csv,
    split_csv_to_dict,
    split_csv_to_set,
    split_csv_to_tuple,
)
from .arg import (
    ARGPARSE_ARG_METADATA_KEY,
    ArgparseArg,
    _eval_type,
    convert_to_command_string,
    make_dataclass_from_func,
)
from .argparse_action import YapxAction
from .exceptions import (
    MutuallyExclusiveArgumentError,
    MutuallyExclusiveRequiredError,
    NoArgsModelError,
    ParserClosedError,
    raise_unsupported_type_error,
)
from .types import Dataclass, NoneType
from .utils import (
    add_argument_to,
    cast_type,
    create_pydantic_model_from_dataclass,
    is_dataclass_type,
    is_pydantic_available,
    is_shtab_available,
    try_isinstance,
    try_issubclass,
)

__all__ = ["ArgumentParser", "run", "run_command"]


if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

if sys.version_info >= (3, 9):
    from typing import _AnnotatedAlias
else:
    from typing_extensions import _AnnotatedAlias

if sys.version_info >= (3, 10):
    from types import UnionType  # pylint: disable=unused-import # noqa: F401

T = TypeVar("T")


class ArgumentParser(argparse.ArgumentParser):
    COMMAND_ATTRIBUTE_NAME: str = "_command"
    FUNC_ATTRIBUTE_NAME: str = "_func"
    ARGS_ATTRIBUTE_NAME: str = "_args_model"

    def __init__(
        self,
        *args: Any,
        prog: Optional[str] = None,
        add_help: bool = True,
        **kwargs: Any,
    ):
        super().__init__(*args, prog=prog, add_help=add_help, **kwargs)

        if is_shtab_available():
            add_argument_to(self, "--print-shell-completion")

        self.kv_separator = "="

        self._mutually_exclusive_args: Dict[
            Optional[Callable[..., Any]],
            Dict[str, List[Tuple[str, Optional[str]]]],
        ] = defaultdict(lambda: defaultdict(list))

        self._inner_type_conversions: Dict[str, Union[type, Callable[[str], Any]]] = {}

    def add_arguments(
        self,
        args_model: Union[Callable[..., Any], Type[Dataclass]],
    ) -> None:
        self._add_arguments(self, args_model)

    def add_command(
        self,
        name: str,
        args_model: Optional[Union[Callable[..., Any], Type[Dataclass]]] = None,
        no_docstring_description: bool = False,
        add_help: bool = True,
        **kwargs: Any,
    ) -> argparse.ArgumentParser:
        subparsers: argparse.Action = self._get_or_add_subparsers()

        # pylint: disable=protected-access
        assert isinstance(subparsers, argparse._SubParsersAction)
        prog: Optional[str] = kwargs.pop("prog", None)
        if prog:
            prog += " " + name

        description_txt: Optional[str] = None
        if not no_docstring_description and args_model:
            description_txt = self._extract_description_from_docstring(args_model)

        # subparsers must populate 'help' in order to
        # be included in shell-completion.
        # ref: https://github.com/iterative/shtab/issues/54#issuecomment-940516963
        parser = subparsers.add_parser(
            name,
            prog=prog,
            add_help=add_help,
            help=description_txt.splitlines()[0] if description_txt else "",
            **kwargs,
        )
        assert isinstance(parser, argparse.ArgumentParser)

        if description_txt:
            self._set_parser_description(parser=parser, description=description_txt)

        if args_model:
            self._add_arguments(parser, args_model)

        return parser

    def _register_funcs(
        self,
        *args: Optional[Callable[..., Any]],
        subparser_kwargs: Optional[Dict[str, Any]] = None,
        no_docstring_description: bool = False,
        **kwargs: Callable[..., Any],
    ) -> None:
        def _register_func(
            func: Callable[..., Any],
            name: Optional[str] = None,
            **sp_kwargs: Any,
        ) -> None:
            name = convert_to_command_string(name if name else func.__name__)
            parser: argparse.ArgumentParser = self.add_command(
                name=name,
                args_model=func,
                no_docstring_description=no_docstring_description,
                **sp_kwargs,
            )
            parser.set_defaults(**{self.FUNC_ATTRIBUTE_NAME: func})

        if not subparser_kwargs:
            subparser_kwargs = {}

        for f in args:
            if f:
                _register_func(f, **subparser_kwargs)

        for nm, f in kwargs.items():
            _register_func(f, name=nm, **subparser_kwargs)

    @staticmethod
    def _is_attribute_inherited(dcls: Type[Dataclass], attr: str) -> bool:
        return attr not in dcls.__annotations__

    @classmethod
    def _add_arguments(
        cls,
        parser: argparse.ArgumentParser,
        args_model: Union[Callable[..., Any], Type[Dataclass]],
    ) -> None:
        """
        Derived from: https://github.com/mivade/argparse_dataclass
        """

        model: Type[Dataclass]

        if is_dataclass_type(args_model):
            model = args_model
        else:
            model = make_dataclass_from_func(args_model)

        if parser.get_default(cls.ARGS_ATTRIBUTE_NAME):
            err: str = "This parser already has args from another dataclass"
            raise ParserClosedError(err)

        parser.set_defaults(**{cls.ARGS_ATTRIBUTE_NAME: model})

        parser_arg_groups: Dict[str, argparse._ArgumentGroup] = {}

        novel_fields: List[Field] = list(
            filter(
                lambda f: not cls._is_attribute_inherited(dcls=model, attr=f.name),
                fields(model),
            ),
        )

        novel_field_args: List[Tuple[Field, ArgparseArg]] = [
            (x, x.metadata.get(ARGPARSE_ARG_METADATA_KEY, ArgparseArg()))
            for x in novel_fields
        ]

        for fld, argparse_argument in novel_field_args:
            fld_type: Union[str, Type[Any]] = fld.type

            # basic support for handling deferred annotation evaluation,
            # when using `from __future__ import annotations`
            if try_isinstance(fld_type, str):
                fld_type = _eval_type(fld_type)
            assert not try_isinstance(fld_type, str)

            if (
                sys.version_info >= (3, 10) and try_isinstance(fld_type, UnionType)
            ) or cls._get_type_origin(fld_type) is Union:
                fld_type = cls._extract_type_from_container(
                    fld_type,
                    is_union_type=True,
                )

            if try_isinstance(fld_type, _AnnotatedAlias):
                for x in cls._get_type_metadata(fld_type):
                    if isinstance(x, Field) and ARGPARSE_ARG_METADATA_KEY in x.metadata:
                        argparse_argument = x.metadata[ARGPARSE_ARG_METADATA_KEY]
                        break

                fld_type = cls._get_type_origin(fld_type)

                if (
                    sys.version_info >= (3, 10) and try_isinstance(fld_type, UnionType)
                ) or cls._get_type_origin(fld_type) is Union:
                    fld_type = cls._extract_type_from_container(
                        fld_type,
                        is_union_type=True,
                    )

            help_type: str = (
                fld_type.__name__
                if try_isinstance(fld_type, type)
                else str(fld_type).rsplit(".", maxsplit=1)[-1]
            )

            argparse_argument.set_dest(fld.name)

            if argparse_argument.required:
                if fld.default is not MISSING:
                    argparse_argument.default = fld.default
                    argparse_argument.required = False
                elif fld.default_factory is not MISSING:
                    argparse_argument.default = fld.default_factory()
                    argparse_argument.required = False

            fld_type_origin: Optional[Type[Any]] = cls._get_type_origin(fld_type)
            if fld_type_origin:
                if fld_type_origin and fld_type_origin is Literal:
                    argparse_argument.choices = list(cls._get_type_args(fld_type))
                    if argparse_argument.choices:
                        fld_type = type(argparse_argument.choices[0])
                else:
                    # this is a type-container (List, Set, Tuple, Dict, ...)
                    if try_issubclass(fld_type_origin, collections.abc.Mapping):
                        fld_type = cls._extract_type_from_container(
                            fld_type,
                            assert_primitive=True,
                        )
                        if not argparse_argument.action:
                            argparse_argument.action = split_csv_to_dict
                    elif (
                        try_issubclass(fld_type_origin, collections.abc.Iterable)
                        or fld_type_origin is set
                    ):
                        fld_type = cls._extract_type_from_container(
                            fld_type,
                            assert_primitive=True,
                        )
                        if not argparse_argument.action:
                            if fld_type_origin is set:
                                argparse_argument.action = split_csv_to_set
                            elif fld_type_origin is tuple:
                                argparse_argument.action = split_csv_to_tuple
                            else:
                                argparse_argument.action = split_csv
                    elif not is_pydantic_available():
                        raise_unsupported_type_error(fld_type)

                    if try_issubclass(fld_type, Enum):
                        argparse_argument.choices = [x.name for x in fld_type]

                    if isinstance(parser, cls):
                        # store desired types for casting later
                        # pylint: disable=protected-access
                        parser._inner_type_conversions[argparse_argument.dest] = (
                            partial(cast_type, target_type=fld_type)
                        )

                    # type-containers must only contain strings
                    # until parsed by argparse.
                    fld_type = str

                    argparse_argument.nargs = "+" if argparse_argument.required else "*"

            elif try_issubclass(fld_type, Enum):
                argparse_argument.choices = list(fld_type)

            argparse_argument.type = partial(cast_type, target_type=fld_type)

            kwargs: Dict[str, Any] = argparse_argument.asdict()
            del kwargs["pos"]

            option_strings: Optional[List[str]] = kwargs.pop("option_strings")
            args: List[str] = (
                []
                if not option_strings
                else [x for x in option_strings if x.startswith(parser.prefix_chars)]
            )

            required: bool = kwargs["required"]

            if not args:
                # positional arg
                del kwargs["required"]
                if not required and not kwargs.get("nargs"):
                    kwargs["nargs"] = "?"

            if fld_type is bool and args:
                for k in "type", "nargs", "const", "choices", "metavar":
                    kwargs.pop(k, None)
                required = False
                kwargs["required"] = False
                if not kwargs["action"]:
                    kwargs["action"] = "store_true"

            # if given `default` cast it to the expected type.
            if kwargs.get("default") is not MISSING:
                if kwargs.get("action") and try_issubclass(
                    kwargs["action"],
                    YapxAction,
                ):
                    # https://stackoverflow.com/a/24448351
                    dummy_namespace: object = type("", (), {})()
                    kwargs["action"](option_strings=args, dest=kwargs["dest"])(
                        parser=parser,
                        namespace=dummy_namespace,
                        values=kwargs["default"],
                    )
                    kwargs["default"] = getattr(dummy_namespace, kwargs["dest"])
                else:
                    kwargs["default"] = argparse_argument.type(kwargs["default"])

            help_msg_parts: List[str] = [f"Type: {help_type}"]

            # pylint: disable=protected-access
            if required:
                help_msg_parts.append("Required")
            else:
                help_default = kwargs.get("default")
                if isinstance(help_default, str):
                    help_default = f"'{help_default}'"
                help_msg_parts.append(f"Default: {help_default}")

            if kwargs.pop("exclusive", False):
                help_msg_parts.append("M.X.")

            if argparse_argument._env_var:
                help_msg_parts.append(f"Env: {argparse_argument._env_var}")

            help_msg: str = f"> {', '.join(help_msg_parts)}"

            if kwargs.get("help"):
                kwargs["help"] += " " + help_msg
            else:
                kwargs["help"] = help_msg

            group: Optional[str] = kwargs.pop("group", None)

            if not group:
                group = "Arguments"

            # if exclusive:
            #     parser_exclusive_args["exclusive"].add_argument(*args, **kwargs)
            # elif group:
            #     arg_group: argparse._ArgumentGroup = parser_arg_groups.get(
            #         group,
            #         parser.add_argument_group(group),
            #     )
            #     arg_group.add_argument(*args, **kwargs)
            #     parser_arg_groups[group] = arg_group
            # elif required:
            #     parser_required_args["required"].add_argument(*args, **kwargs)
            # else:
            #     parser_optional_args["optional"].add_argument(*args, **kwargs)

            arg_group: argparse._ArgumentGroup = parser_arg_groups.get(group, None)

            if not arg_group:
                arg_group = parser.add_argument_group(group)
                parser_arg_groups[group] = arg_group

            if argparse_argument.exclusive:
                if required:
                    err: str = (
                        "A mutually-exclusive argument cannot be required:"
                        f" {argparse_argument.dest}"
                    )
                    raise MutuallyExclusiveRequiredError(err)

                parser._mutually_exclusive_args[
                    parser._defaults.get(cls.FUNC_ATTRIBUTE_NAME)
                ][group].append(
                    (
                        argparse_argument.dest,
                        args[-1] if args else None,
                    ),
                )

            arg_group.add_argument(*args, **kwargs)

    @staticmethod
    def _get_type_origin(t: Type[Any]) -> Optional[Type[Any]]:
        return getattr(t, "__origin__", None)

    @staticmethod
    def _get_type_args(t: Type[Any]) -> Tuple[Type[Any], ...]:
        return getattr(t, "__args__", ())

    @staticmethod
    def _get_type_metadata(t: Type[Any]) -> Tuple[Type[Any], ...]:
        return getattr(t, "__metadata__", ())

    @classmethod
    def _extract_type_from_container(
        cls,
        type_container: Type[Any],
        assert_primitive: bool = False,
        is_union_type: bool = False,
    ) -> Type[Any]:
        type_container_origin: Any = (
            Union if is_union_type else cls._get_type_origin(type_container)
        )

        if type_container_origin is None:
            raise TypeError(
                f"Given type is not a container: {type_container.__name__}",
            )

        type_container_args = cls._get_type_args(type_container)

        results: List[Type[Any]] = [
            a for a in type_container_args if a is not ... and a is not NoneType
        ]

        type_container_subtype: type

        if (
            type_container_origin is Union
            or try_issubclass(type_container_origin, collections.abc.Sequence)
            or type_container_origin is set
        ):
            if len(results) != 1:
                results.clear()
            else:
                type_container_subtype = results[0]
        elif try_issubclass(type_container_origin, collections.abc.Mapping):
            if len(results) != 2:
                results.clear()
            else:
                key_type, value_type = results
                if key_type is not str:
                    raise TypeError("Dictionary keys must be type `str`")
                type_container_subtype = value_type
        elif not is_pydantic_available():
            raise_unsupported_type_error(type_container)

        if not results:
            raise TypeError("Too many types in container")

        type_container_subtype_origin = cls._get_type_origin(type_container_subtype)

        if type_container_subtype_origin is Union:
            type_container_subtype = cls._extract_type_from_container(
                type_container_subtype,
            )
            type_container_subtype_origin = cls._get_type_origin(type_container_subtype)

        if (
            assert_primitive
            and type_container_subtype_origin
            and not is_pydantic_available()
        ):
            raise_unsupported_type_error(type_container_subtype_origin)

        return type_container_subtype

    def _get_subparsers(self) -> Optional[argparse._SubParsersAction]:
        for a in self._actions:
            if isinstance(
                a,
                argparse._SubParsersAction,  # pylint: disable=protected-access
            ):
                return a
        return None

    def _get_or_add_subparsers(self) -> argparse._SubParsersAction:
        subparsers = self._get_subparsers()

        if not subparsers:
            subparsers = self.add_subparsers(dest=self.COMMAND_ATTRIBUTE_NAME)

        return subparsers

    def _post_parse_args(  # type: ignore[override]
        self,
        namespace: argparse.Namespace,
    ) -> argparse.Namespace:
        # delete vars created by shtab
        sh_complete_attr = "print_shell_completion"
        if hasattr(namespace, sh_complete_attr):
            delattr(namespace, sh_complete_attr)

        func_mx_arg_groups: Dict[str, List[Tuple[str, Optional[str]]]] = (
            self._mutually_exclusive_args.get(
                getattr(namespace, self.FUNC_ATTRIBUTE_NAME, None),
                {},
            )
        )
        for g_args in func_mx_arg_groups.values():
            # argparse will always return a list when the argument type is `list | None`,
            # so we exclude both null and empty values
            # e.g.,: parser = argparse.ArgumentParser(); g = parser.add_mutually_exclusive_group(); g.add_argument("test", nargs="*", default=None); g.add_argument("--test2", default=None); parser.parse_args(["--test2", "aye"])
            mx_flags_found: List[str] = [
                flag if flag else dest
                for dest, flag in g_args
                if hasattr(namespace, dest)
                for attr_value in [getattr(namespace, dest)]
                if (
                    (
                        try_isinstance(attr_value, str)
                        or not try_isinstance(
                            attr_value,
                            collections.abc.Iterable,
                        )
                    )
                    and attr_value is not None
                )
                or (
                    not try_isinstance(attr_value, str)
                    and try_isinstance(
                        attr_value,
                        collections.abc.Iterable,
                    )
                    and attr_value
                )
            ]

            if len(mx_flags_found) > 1:
                raise MutuallyExclusiveArgumentError(", ".join(mx_flags_found))

        return namespace

    def parse_args(  # type: ignore[override]
        self,
        args: Optional[Sequence[str]] = None,
        namespace: Optional[argparse.Namespace] = None,
    ) -> argparse.Namespace:
        parsed: argparse.Namespace = super().parse_args(args=args, namespace=namespace)
        return self._post_parse_args(parsed)

    def parse_known_args(  # type: ignore[override]
        self,
        args: Optional[Sequence[str]] = None,
        namespace: Optional[argparse.Namespace] = None,
    ) -> Tuple[argparse.Namespace, List[str]]:
        parsed: argparse.Namespace
        unknown: List[str]
        parsed, unknown = super().parse_known_args(args=args, namespace=namespace)
        return self._post_parse_args(parsed), unknown

    def parse_known_args_to_model(
        self,
        args: Optional[Sequence[str]] = None,
        args_model: Optional[Type[Dataclass]] = None,
        skip_pydantic_validation: bool = False,
    ) -> Tuple[Dataclass, List[str]]:
        parsed_args: argparse.Namespace
        unknown_args: List[str]
        parsed_args, unknown_args = self.parse_known_args(args)

        return (
            self._parse_args_to_model(
                args=parsed_args,
                args_model=args_model,
                skip_pydantic_validation=skip_pydantic_validation,
            ),
            unknown_args,
        )

    def parse_args_to_model(
        self,
        args: Optional[Sequence[str]] = None,
        args_model: Optional[Type[Dataclass]] = None,
        skip_pydantic_validation: bool = False,
    ) -> Dataclass:
        parsed_args: argparse.Namespace = self.parse_args(args)
        return self._parse_args_to_model(
            args=parsed_args,
            args_model=args_model,
            skip_pydantic_validation=skip_pydantic_validation,
        )

    def _parse_args_to_model(
        self,
        args: argparse.Namespace,
        args_model: Optional[Type[Dataclass]] = None,
        skip_pydantic_validation: bool = False,
    ) -> Dataclass:
        parsed_args: Dict[str, Any] = vars(args)

        if not args_model:
            args_model = parsed_args.get(self.ARGS_ATTRIBUTE_NAME)
            if not args_model:
                raise NoArgsModelError

        args_union: Dict[str, Any] = self._union_args_with_model(
            args_dict=parsed_args,
            args_model=args_model,
        )

        if not skip_pydantic_validation and is_pydantic_available():
            args_union = vars(
                create_pydantic_model_from_dataclass(args_model)(**args_union),
            )

        return args_model(**args_union)

    @staticmethod
    def _union_args_with_model(
        args_dict: Dict[str, Any],
        args_model: Type[Dataclass],
    ) -> Dict[str, Any]:
        return {k: v for k, v in args_dict.items() if k in args_model.__annotations__}

    def print_help(
        self,
        file: Optional[IO[str]] = None,
        full: bool = False,
    ) -> None:
        super().print_help(file)

        if full:
            subparsers: Optional[argparse._SubParsersAction] = self._get_subparsers()
            separator: str = "\n" + ("*" * 80)
            if subparsers:
                for choice, subparser in subparsers.choices.items():
                    print(separator)
                    print(f">>> {choice}")
                    subparser.print_help(file)

    @classmethod
    def _docstring_to_description(
        cls,
        parser: argparse.ArgumentParser,
        args_model: Union[Callable[..., Any], Type[Dataclass]],
    ) -> None:
        cls._set_parser_description(
            parser,
            description=cls._extract_description_from_docstring(args_model),
        )

    @classmethod
    def _set_parser_description(
        cls,
        parser: argparse.ArgumentParser,
        description: Optional[str],
    ):
        if not description:
            return

        parser.description = description

        if "\n" in description:
            # allow newlines in parser description
            parser.formatter_class = argparse.RawTextHelpFormatter

    @staticmethod
    def _extract_description_from_docstring(
        args_model: Union[Callable[..., Any], Type[Dataclass]],
    ) -> Optional[str]:
        description_lines: Optional[List[str]] = None

        if args_model.__doc__ and not args_model.__doc__.startswith(
            args_model.__name__ + "(",
        ):
            description_lines = args_model.__doc__.strip().splitlines()

        if not description_lines:
            return None

        text_block_ends_at: int = 0
        for i, line in enumerate(description_lines):
            if not line.strip():
                text_block_ends_at = i
                break

        if text_block_ends_at > 0:
            description_lines = description_lines[:text_block_ends_at]

        return "\n".join(x.strip() for x in description_lines)

    @staticmethod
    def _run_func(
        parser: "ArgumentParser",
        func: Callable[..., Any],
        args_model: Type[Any],
        args: List[str],
        linked_func: Optional[Callable[..., Any]] = None,
        relay_value: Optional[Any] = None,
    ) -> Any:
        func_args: List[str] = []
        func_kwargs: Dict[str, Optional[Any]] = {}

        extra_args_ok: bool = False
        accepts_kwargs: bool = False
        accepts_pos_args: bool = False
        accepts_extra_args: bool = False
        accepts_relay_value: bool = False

        for p in signature(func).parameters.values():
            if str(p).startswith("**"):
                accepts_kwargs = True
            elif str(p).startswith("*"):
                accepts_pos_args = True
            elif p.name == "_extra_args":
                accepts_extra_args = True
            elif p.name == "_relay_value":
                accepts_relay_value = True

        extra_args_ok = accepts_pos_args or accepts_kwargs or accepts_extra_args

        if not extra_args_ok and linked_func:
            extra_args_ok = any(
                str(p).startswith("*") or p.name == "_extra_args"
                for p in signature(linked_func).parameters.values()
            )

        if accepts_relay_value:
            func_kwargs["_relay_value"] = relay_value

        model_inst: Dataclass
        if extra_args_ok:
            unknown_args: List[str]
            model_inst, unknown_args = parser.parse_known_args_to_model(
                args=args,
                args_model=args_model,
            )

            if accepts_extra_args:
                func_kwargs["_extra_args"] = unknown_args

            if accepts_pos_args:
                func_args = unknown_args

            if accepts_kwargs:
                func_kwargs.update(
                    _split_csv_to_dict(
                        unknown_args,
                        kv_separator=parser.kv_separator,
                    ),
                )
        else:
            model_inst = parser.parse_args_to_model(
                args=args,
                args_model=args_model,
            )

        return func(*func_args, **vars(model_inst), **func_kwargs)

    @classmethod
    def _run(
        cls,
        *args: Optional[Callable[..., Any]],
        _args: Optional[List[str]] = None,
        _prog: Optional[str] = None,
        _no_help: bool = False,
        _print_help: bool = False,
        _no_docstring_description: bool = False,
        **kwargs: Callable[..., Any],
    ) -> Any:
        parser_shared_kwargs: Dict[str, Any] = {"prog": _prog, "add_help": not _no_help}

        parser: ArgumentParser = cls(**parser_shared_kwargs)

        parser.set_defaults(
            **{cls.ARGS_ATTRIBUTE_NAME: None, cls.FUNC_ATTRIBUTE_NAME: None},
        )

        setup_func: Optional[Callable[..., Any]] = None
        setup_func_arg_model: Optional[Type[Any]] = None
        cmd_funcs: List[Callable[..., Any]] = []

        if args:
            setup_func = args[0]

            if setup_func:
                setup_func_arg_model = make_dataclass_from_func(setup_func)
                parser.add_arguments(setup_func_arg_model)
                if not _no_docstring_description:
                    cls._docstring_to_description(
                        parser=parser,
                        args_model=setup_func_arg_model,
                    )

            _cmd_funcs: Iterable[Optional[Callable[..., Any]]] = args[1:]
            assert isinstance(_cmd_funcs, tuple)
            cmd_funcs.extend(_cmd_funcs)

        parser._register_funcs(
            *cmd_funcs,
            subparser_kwargs=parser_shared_kwargs,
            no_docstring_description=_no_docstring_description,
            **kwargs,
        )

        if _args is None:
            _args = sys.argv[1:]

        if _print_help or "--help-full" in _args:
            parser.print_help(full=True)
            parser.exit()

        known_args: argparse.Namespace
        known_args, _ = parser.parse_known_args(_args)
        parsed_args: Dict[str, Any] = vars(known_args)

        # parsed_args.get(cls.COMMAND_ATTRIBUTE_NAME)

        func: Optional[Callable[..., Any]] = parsed_args.get(cls.FUNC_ATTRIBUTE_NAME)
        args_model: Optional[Type[Any]] = parsed_args.get(cls.ARGS_ATTRIBUTE_NAME)
        relay_value: Any = None

        setup_result: Any = None
        if setup_func:
            assert setup_func_arg_model
            setup_result = cls._run_func(
                parser=parser,
                func=setup_func,
                args_model=setup_func_arg_model,
                linked_func=func,
                args=_args,
            )

        if try_isinstance(setup_result, GeneratorType):
            try:
                relay_value = next(setup_result)
            except StopIteration:
                relay_value = setup_result
        else:
            relay_value = setup_result

        try:
            if func:
                relay_value = cls._run_func(
                    parser=parser,
                    func=func,
                    args_model=args_model,
                    args=_args,
                    linked_func=setup_func,
                    relay_value=relay_value,
                )
        finally:
            if try_isinstance(setup_result, GeneratorType):
                for gen_result in setup_result:
                    if not func:
                        relay_value = gen_result

        return relay_value


def run(
    *args: Optional[Callable[..., Any]],
    _args: Optional[List[str]] = None,
    _prog: Optional[str] = None,
    _no_help: bool = False,
    _print_help: bool = False,
    _no_docstring_description: bool = False,
    **kwargs: Callable[..., Any],
) -> Any:
    # pylint: disable=protected-access
    return ArgumentParser._run(
        *args,
        _args=_args,
        _prog=_prog,
        _no_help=_no_help,
        _print_help=_print_help,
        _no_docstring_description=_no_docstring_description,
        **kwargs,
    )


@wraps(run)
def run_command(*args, **kwargs) -> Any:
    return run(
        None,
        *args,
        **kwargs,
    )



import argparse
from dataclasses import dataclass
from typing import Any, Dict

import yapx


def test_add_command() -> None:
    # 1. ARRANGE
    @dataclass
    class RootModel:
        debug: bool

    @dataclass
    class CmdModel(RootModel):
        str_value: str
        int_value: int
        bool_value: bool

    expected_cmd_name: str = "my-command"

    expected: Dict[str, Any] = {
        "debug": True,
        yapx.ArgumentParser.COMMAND_ATTRIBUTE_NAME: expected_cmd_name,
        yapx.ArgumentParser.ARGS_ATTRIBUTE_NAME: CmdModel,
        "str_value": "abc",
        "int_value": 5,
        "bool_value": True,
    }

    cli_args = [
        "--debug",
        expected_cmd_name,
        "--str-value",
        "abc",
        "--int-value",
        "5",
        "--bool-value",
    ]

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(RootModel)
    cmd_parser: argparse.ArgumentParser = parser.add_command(
        name=expected_cmd_name,
        args_model=CmdModel,
    )

    args: Dict[str, Any] = vars(parser.parse_args(cli_args))

    # 3. ASSERT
    # pylint: disable=protected-access
    assert cmd_parser
    assert expected_cmd_name in parser._get_or_add_subparsers().choices
    assert args == expected



from dataclasses import dataclass
from typing import Optional

import pytest

import yapx


@pytest.mark.parametrize(
    "x",
    [-1, 1, "1", "true", "t", "yes", "y", "on", True, ["hello"]],
)
def test_cast_bool_true(x):
    assert yapx.utils.cast_bool(x) is True


@pytest.mark.parametrize("x", [0, "0", "false", "f", "no", "n", "off", None, False, []])
def test_cast_bool_false(x):
    assert yapx.utils.cast_bool(x) is False


def test_is_dataclass_type():
    @dataclass
    class Something:
        nothing: Optional[str] = None

    assert yapx.utils.is_dataclass_type(Something)
    assert yapx.utils.is_dataclass_type(Something())


def test_try_isinstance():
    assert yapx.utils.try_isinstance("", str)
    assert not yapx.utils.try_isinstance(0, str)
    assert not yapx.utils.try_isinstance(None, str)


def test_try_issubclass():
    class Something(str):
        ...

    assert yapx.utils.try_issubclass(Something, str)
    assert not yapx.utils.try_issubclass(Something(), str)
    assert not yapx.utils.try_issubclass(None, str)


def test_coalesce():
    assert yapx.utils.coalesce("", "this") == ""
    assert yapx.utils.coalesce("", "this", null_or_empty=True) == "this"
    assert yapx.utils.coalesce(None, "this") == "this"



import sys
from enum import Enum
from typing import Any, Dict, Optional, Sequence, Union

if sys.version_info >= (3, 8):
    from typing import Literal  # pylint: disable=unused-import # noqa: F401
else:
    from typing_extensions import Literal  # pylint: disable=unused-import # noqa: F401

if sys.version_info >= (3, 9):
    from typing import Annotated  # pylint: disable=unused-import # noqa: F401
else:
    from typing_extensions import (  # pylint: disable=unused-import # noqa: F401
        Annotated,
    )

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

NoneType = type(None)

ArgValueType = Union[None, str, Sequence[str]]

__all__ = [
    "Dataclass",
    "ArgumentParser",
    "NoneType",
    "ArgValueType",
    "Literal",
    "Enum",
    "Annotated",
    "Protocol",
]


class Dataclass(Protocol):
    __dataclass_fields__: Any


class ArgumentParser(Protocol):
    kv_separator = str

    _mutually_exclusive_args: Dict[str, Dict[str, Optional[str]]]

    _inner_type_conversions: Dict[str, type]

    def print_help(self) -> None:
        ...

    def exit(self, status: int = 0, message: Optional[str] = None) -> None:
        ...



from __future__ import annotations

import yapx


def test_add_arguments_deferred():
    # 1. ARRANGE

    # pylint: disable=unused-argument
    # noinspection Annotator
    def func(
        value: None | list[str] = yapx.arg(lambda: ["hello world"]),
        value_kv: None | dict[str, str] = yapx.arg(None),
    ):
        ...

    cli_args: list[str] = ["--value-kv", "hello=world", "this=that"]

    expected_value: list[str] = ["hello world"]
    expected_value_kv: dict[str, str] = {"hello": "world", "this": "that"}

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(func)
    args: type[yapx.types.Dataclass] = parser.parse_args_to_model(cli_args)

    # 3. ASSERT
    assert args
    assert hasattr(args, "value")
    assert args.value == expected_value
    assert hasattr(args, "value_kv")
    assert args.value_kv == expected_value_kv



import os
import sys
from argparse import Action
from contextlib import suppress
from dataclasses import MISSING, Field, dataclass, field, make_dataclass
from inspect import Parameter, _empty, signature
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    get_type_hints,
)

import yapx  # pylint: disable=unused-import # noqa: F401

from .types import Dataclass

if sys.version_info >= (3, 8):
    from typing import Literal  # pylint: disable=unused-import # noqa: F401
else:
    from typing_extensions import Literal  # pylint: disable=unused-import # noqa: F401

if sys.version_info >= (3, 9):
    from typing import Annotated  # pylint: disable=unused-import # noqa: F401
else:
    from typing_extensions import (  # pylint: disable=unused-import # noqa: F401
        Annotated,
    )


__all__ = ["arg"]


ARGPARSE_ARG_METADATA_KEY: str = "_argparse_argument"


@dataclass
class ArgparseArg:
    dest: Optional[str] = None
    option_strings: Optional[Sequence[str]] = None
    type: Union[None, Type[Any], Callable[[str], Any]] = None
    action: Union[None, str, Type[Action]] = None
    required: bool = True
    group: Optional[str] = None
    exclusive: Optional[bool] = False
    nargs: Optional[str] = None
    const: Optional[Any] = None
    default: Optional[Any] = None
    choices: Optional[Sequence[Any]] = None
    help: Optional[str] = None
    metavar: Optional[str] = None
    pos: Optional[bool] = False
    _env_var: Union[None, str, Sequence[str]] = None

    def __post_init__(self) -> None:
        self.set_dest(self.dest)

    def set_dest(self, value: Optional[str]) -> None:
        self.dest = value

        if self.option_strings:
            if self.pos:
                self.option_strings = None
            elif isinstance(self.option_strings, str):
                self.option_strings = [self.option_strings]
        elif self.dest and not self.pos:
            self.option_strings = [convert_to_flag_string(self.dest)]

    def asdict(self) -> Dict[str, Any]:
        return {k: v for k, v in vars(self).items() if not k.startswith("_")}


def arg(
    default: Optional[Any] = MISSING,
    env: Union[None, str, Sequence[str]] = None,
    pos: Optional[bool] = False,
    group: Optional[str] = None,
    exclusive: Optional[bool] = False,
    flags: Union[None, str, Sequence[str]] = None,
    # pylint: disable=redefined-builtin
    help: Optional[str] = None,
    metavar: Optional[str] = None,
    action: Union[None, str, Type[Action]] = None,
) -> Field:
    if env:
        if isinstance(env, str):
            env = [env]
        for e in env:
            value_from_env = os.getenv(e, None)
            if value_from_env:
                default = value_from_env
                break

            env_file = os.getenv(e + "_FILE", None)
            if env_file and os.path.exists(env_file):
                with open(env_file, encoding="utf8") as f:
                    value_from_file = f.read().strip()
                if value_from_file:
                    default = value_from_file
                    break

    metadata: Dict[str, ArgparseArg] = {
        ARGPARSE_ARG_METADATA_KEY: ArgparseArg(
            action=action,
            pos=pos,
            option_strings=flags,
            required=bool(default is MISSING),
            default=(
                None
                if default is MISSING
                else default() if callable(default) else default
            ),
            group=group,
            exclusive=exclusive,
            help=help,
            metavar=metavar,
            _env_var=env,
        ),
    }

    kwargs: Dict[str, Any] = {"metadata": metadata}

    default_param: str = "default_factory" if callable(default) else "default"
    kwargs[default_param] = default

    return field(**kwargs)


def convert_to_command_string(x: str) -> str:
    x = x.strip()

    x_prefix: str = "x_"
    under: str = "_"
    if x.startswith(x_prefix):
        x = x[len(x_prefix) :]
    elif x.startswith(under):
        # `_xxx_cmd_name` --> `cmd-name`
        next_under: int = 0
        with suppress(ValueError):
            next_under = x.index(under, 1)
        if not next_under:
            x = x.lstrip(under)
        else:
            x = x[next_under + 1 :]

    x = x.strip("-").lower().replace("_", "-")

    if not x:
        raise ValueError("Expected at least one character")

    return x


def convert_to_flag_string(x: str) -> str:
    return "--" + convert_to_command_string(x)


def convert_to_short_flag_string(x: str) -> str:
    return "-" + convert_to_command_string(x)[0]


def _eval_type(type_str: str) -> Type[Any]:
    if "[" in type_str:
        # None | list[str] --> None | List[str]
        type_str = "|".join(
            (y.capitalize() if "[" in y else y)
            for x in type_str.split("|")
            for y in [x.strip()]
        )
    if "|" in type_str:
        # None | str --> Union[None, str]
        type_str = f"Union[{type_str.replace('|', ',')}]"

    type_obj: Type[Any] = eval(type_str)  # pylint: disable=eval-used

    return type_obj


def make_dataclass_from_func(
    func: Callable[..., Any],
    base_classes: Optional[Tuple[Type[Dataclass], ...]] = None,
    include_private_params: bool = False,
) -> Type[Dataclass]:
    if base_classes is None:
        base_classes = ()

    fields: List[Tuple[str, Type[Any], Field]] = []

    func_signature = signature(func)
    type_hints: Dict[str, Any]
    try:
        include_extras: Dict[str, bool] = {}
        if sys.version_info >= (3, 9):
            include_extras["include_extras"] = True
        type_hints = get_type_hints(func, **include_extras)
    except (TypeError, NameError):
        # this can happen if deferred evaluation is used,
        # via `from __future__ import annotations`
        type_hints = {}

    param: Parameter
    for param in func_signature.parameters.values():
        if str(param).startswith("*") or (
            not include_private_params and param.name.startswith("_")
        ):
            continue

        annotation: Type[Any] = (
            _eval_type(param.annotation)
            if isinstance(param.annotation, str)
            else type_hints.get(param.name, param.annotation)
        )

        default: Any = param.default
        default_value: Any

        field_metadata: Field

        if isinstance(default, Field):
            if default.default_factory is not MISSING:
                default_value = default.default_factory()
            else:
                default_value = default.default
            field_metadata = default
        else:
            if default is _empty:  # pylint: disable=protected-access
                default_value = MISSING
            else:
                default_value = default

            field_metadata = arg(default=default_value)

        fallback_type: Type[Any] = str

        # pylint: disable=protected-access
        if annotation is _empty:
            if default_value is MISSING:
                annotation = fallback_type
            elif default_value is None:
                annotation = Optional[fallback_type]
            else:
                type_of_default: Type[Any] = type(default_value)
                if type_of_default in (str, int, float, bool):
                    annotation = type_of_default
                else:
                    raise TypeError(
                        (
                            f"Provide explicit type annotation for '{param.name}' "
                            f"(type  of default is '{type_of_default})'"
                        ),
                    )

        fields.append((param.name, annotation, field_metadata))

    dc: Type[Dataclass] = make_dataclass(
        cls_name="Dataclass_" + func.__name__,
        bases=base_classes,
        fields=fields,
    )

    dc.__doc__ = func.__doc__

    return dc



import collections.abc
import shlex
from typing import Any, Dict, List, Optional, Set, Tuple, Type, TypeVar

from .argparse_action import YapxAction, argparse_action
from .types import ArgumentParser, ArgValueType
from .utils import coalesce, try_isinstance, try_issubclass

# @argparse_action
# # pylint: disable=unused-argument
# def to_str(
#     value: ArgValueType,
#     *,
#     action: argparse.Action,
#     parser: argparse.ArgumentParser,
#     namespace: argparse.Namespace,
#     option_string: Optional[str],
# ) -> Optional[str]:
#     return _cast(value, to=str)


T = TypeVar("T", bound=type)


def _split_csv_sequence(
    values: ArgValueType,
    target_type: T,
) -> Optional[List[T]]:
    if isinstance(values, str) and values:
        values = [values]
    elif (
        values is None
        or not try_issubclass(type(values), collections.abc.Sequence)
        or len(values) == 0
        or not isinstance(values[0], str)
    ):
        return values

    all_values: List[str] = []

    list_prefix: str = "list["
    list_suffix: str = "]"

    if all(x.endswith(",") for x in values[:-1]):
        values = [x.rstrip(",") for x in values]

    for value in values:
        value_clean: str = value.strip()

        if (
            value_clean
            and value_clean.lower().startswith(list_prefix)
            and value_clean.lower().endswith(list_suffix)
        ):
            all_values.extend(
                _split_csv_sequence(
                    values=shlex.split(
                        value_clean[len(list_prefix) : -len(list_suffix)].strip(),
                    ),
                    target_type=target_type,
                ),
            )
        else:
            all_values.append(target_type(value))

    return all_values


@argparse_action
def split_csv(
    values: ArgValueType,
    *,
    action: YapxAction,
    parser: ArgumentParser,
    **_kwargs: Any,
) -> Optional[List[Optional[Any]]]:
    return _split_csv_sequence(
        values,
        target_type=parser._inner_type_conversions.get(  # pylint: disable=protected-access
            action.dest,
        ),
    )


@argparse_action
def split_csv_to_tuple(
    values: ArgValueType,
    *,
    action: YapxAction,
    parser: ArgumentParser,
    **_kwargs: Any,
) -> Optional[Tuple[Optional[Any], ...]]:
    split_values: Optional[List[Optional[Any]]] = _split_csv_sequence(
        values,
        target_type=parser._inner_type_conversions.get(  # pylint: disable=protected-access
            action.dest,
        ),
    )

    if split_values is None:
        return None

    return tuple(split_values)


@argparse_action
def split_csv_to_set(
    values: ArgValueType,
    *,
    action: YapxAction,
    parser: ArgumentParser,
    **_kwargs: Any,
) -> Optional[Set[Optional[Any]]]:
    if try_isinstance(values, set):
        return values

    split_values: Optional[List[Optional[Any]]] = _split_csv_sequence(
        values,
        target_type=parser._inner_type_conversions.get(  # pylint: disable=protected-access
            action.dest,
        ),
    )
    if split_values is None:
        return None

    return set(split_values)


def _split_csv_to_dict(
    values: ArgValueType,
    kv_separator: str,
    target_value_type: Optional[Type[Any]] = None,
) -> Optional[Dict[str, Optional[Any]]]:
    if try_isinstance(values, dict):
        return values

    split_values: Optional[List[Optional[Any]]] = _split_csv_sequence(
        values,
        target_type=str,
    )

    if split_values is None:
        return None

    return {
        x_split[0].strip(): (
            v if v is None or target_value_type is None else target_value_type(v)
        )
        for x in split_values
        if x
        for x_split in [x.split(kv_separator, maxsplit=1)]
        for v in [
            (
                None
                if len(x_split) < 2
                else coalesce(x_split[1].strip(), None, null_or_empty=True)
            ),
        ]
    }


@argparse_action
def split_csv_to_dict(
    values: ArgValueType,
    *,
    action: YapxAction,
    parser: ArgumentParser,
    **_kwargs: Any,
) -> Optional[Dict[str, Optional[Any]]]:
    return _split_csv_to_dict(
        values,
        kv_separator=parser.kv_separator,
        target_value_type=parser._inner_type_conversions.get(  # pylint: disable=protected-access
            action.dest,
            str,
        ),
    )



from argparse import Action
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Set, Tuple

import pytest
from _pytest.capture import CaptureFixture, CaptureResult

import yapx
from yapx.argparse_action import YapxAction


def test_split_csv():
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        values: List[Optional[str]]

    expected: List[Optional[str]] = [
        "1",
        "1",
        "1",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
    ]
    cli_args = ["--values"] + expected

    expected_action_name: str = "split_csv"

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)

    args: Dict[str, Any] = vars(parser.parse_args(cli_args))

    # 3. ASSERT
    # pylint: disable=protected-access
    parser_action: Action = parser._actions[2]
    assert isinstance(parser_action, YapxAction)
    # pylint: disable=protected-access
    assert parser_action.name == expected_action_name

    assert "values" in args
    assert args["values"] == expected


def test_split_csv_to_set():
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        values: Set[Optional[str]]

    expected: Set[Optional[str]] = {
        "",
        "1",
        "2",
        "3  4",
        "5",
        "6  7",
        "8",
        "9",
    }
    cli_args = ["--values"] + list(expected)

    expected_action_name: str = "split_csv_to_set"

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)

    args: Dict[str, Any] = vars(parser.parse_args(cli_args))

    # 3. ASSERT
    # pylint: disable=protected-access
    parser_action: Action = parser._actions[2]
    assert isinstance(parser_action, YapxAction)
    # pylint: disable=protected-access
    assert parser_action.name == expected_action_name

    assert "values" in args
    assert args["values"] == expected


def test_split_csv_to_tuple():
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        values: Tuple[Optional[str], ...]

    expected: Tuple[Optional[str]] = (
        "",
        "",
        "1",
        "1",
        "1",
        "1",
        "2",
        "3  4",
        "5",
        "6  7",
        "8",
        "9",
        "",
    )
    cli_args = ["--values"] + list(expected)

    expected_action_name: str = "split_csv_to_tuple"

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)

    args: Dict[str, Any] = vars(parser.parse_args(cli_args))

    # 3. ASSERT
    # pylint: disable=protected-access
    parser_action: Action = parser._actions[2]
    assert isinstance(parser_action, YapxAction)
    # pylint: disable=protected-access
    assert parser_action.name == expected_action_name

    assert "values" in args
    assert args["values"] == expected


def test_split_csv_to_dict():
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        values: Dict[str, Optional[str]]

    expected: Dict[str, Optional[str]] = {
        "1": None,
        "2": None,
        "3": None,
        "4": "hello",
        "5": None,
        "6": "world",
        "7": None,
        "8 9": None,
    }
    cli_args = [
        "--values",
        "1,",
        "1,",
        "1,",
        "1,",
        "2,",
        "3,",
        " 4 = hello,",
        "5,",
        "6  = world,",
        " 7 = ,",
        " 8 9 ,",
    ]

    expected_action_name: str = "split_csv_to_dict"

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)

    args: Dict[str, Any] = vars(parser.parse_args(cli_args))

    # 3. ASSERT
    # pylint: disable=protected-access
    parser_action: Action = parser._actions[2]
    assert isinstance(parser_action, YapxAction)
    # pylint: disable=protected-access
    assert parser_action.name == expected_action_name

    assert "values" in args
    assert args["values"] == expected


def test_print_help_full(capsys: CaptureFixture):
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        values: str = yapx.arg(env="TEST_VALUES")

    cli_args: List[str] = ["-h", "--values", "1"]

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)
    parser.add_command("subcmd1", ArgsModel)
    parser.add_command("subcmd2", ArgsModel)
    with pytest.raises(SystemExit) as pexit:
        parser.parse_args(cli_args)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out
    assert pexit.value.code == 0


def test_print_help_subparser(capsys: CaptureFixture):
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        values: str = yapx.arg(env="TEST_VALUES")

    expected_cmd: str = "subcmd1"

    cli_args: List[str] = [expected_cmd, "-h"]

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)
    parser.add_command(expected_cmd, ArgsModel)
    parser.add_command("subcmd2", ArgsModel)
    with pytest.raises(SystemExit) as pexit:
        parser.parse_args(cli_args)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert expected_cmd in captured.out
    assert pexit.value.code == 0


def test_print_docstring_help(capsys: CaptureFixture):
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        """hello world"""

        values: str = yapx.arg(env="TEST_VALUES")

    expected_cmd: str = "subcmd1"
    expected_txt: List[str] = [
        "{subcmd1,subcmd2}",
        "--values",
        "--help",
        "hello world",
        "TEST_VALUES",
    ]

    cli_args: List[str] = ["-h"]

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)
    parser.add_command(expected_cmd, ArgsModel)
    parser.add_command("subcmd2", ArgsModel)
    with pytest.raises(SystemExit) as pexit:
        parser.parse_args(cli_args)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert expected_cmd in captured.out
    assert pexit.value.code == 0

    for e in expected_txt:
        assert e in captured.out



import sys
from argparse import Action, ArgumentError
from dataclasses import MISSING, dataclass, make_dataclass
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Tuple, Type

import pytest

import yapx
from yapx.arg import ArgparseArg
from yapx.exceptions import ParserClosedError

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

if sys.version_info >= (3, 9):
    from typing import Annotated, _AnnotatedAlias
else:
    from typing_extensions import Annotated, _AnnotatedAlias


@pytest.mark.parametrize(
    ("name", "type_annotation", "default", "expected"),
    [
        # str, annotated
        (
            "a_str",
            Annotated[str, "nothingburger"],
            MISSING,
            ArgparseArg(
                option_strings=["--a-str"],
                help="> Type: str, Required",
                required=True,
            ),
        ),
        (
            "a_str",
            Annotated[str, yapx.arg("hello")],
            MISSING,
            ArgparseArg(
                option_strings=["--a-str"],
                help="> Type: str, Default: 'hello'",
                required=False,
                default="hello",
            ),
        ),
        (
            "a_str",
            Annotated[Optional[str], yapx.arg("hello")],
            None,
            ArgparseArg(
                option_strings=["--a-str"],
                help="> Type: str, Default: 'hello'",
                required=False,
                default="hello",
            ),
        ),
        # str
        (
            "a_str",
            str,
            MISSING,
            ArgparseArg(
                option_strings=["--a-str"],
                help="> Type: str, Required",
                required=True,
            ),
        ),
        (
            "a_str",
            str,
            "hello",
            ArgparseArg(
                option_strings=["--a-str"],
                help="> Type: str, Default: 'hello'",
                required=False,
                default="hello",
            ),
        ),
        (
            "a_str",
            Optional[str],
            None,
            ArgparseArg(
                option_strings=["--a-str"],
                help="> Type: str, Default: None",
                required=False,
                default=None,
            ),
        ),
        # int
        (
            "a_int",
            int,
            MISSING,
            ArgparseArg(
                option_strings=["--a-int"],
                help="> Type: int, Required",
                required=True,
            ),
        ),
        (
            "a_int",
            int,
            123,
            ArgparseArg(
                option_strings=["--a-int"],
                help="> Type: int, Default: 123",
                required=False,
                default=123,
            ),
        ),
        (
            "a_int",
            Optional[int],
            None,
            ArgparseArg(
                option_strings=["--a-int"],
                help="> Type: int, Default: None",
                required=False,
                default=None,
            ),
        ),
        # float
        (
            "a_float",
            float,
            MISSING,
            ArgparseArg(
                option_strings=["--a-float"],
                help="> Type: float, Required",
                required=True,
            ),
        ),
        (
            "a_float",
            float,
            3.14,
            ArgparseArg(
                option_strings=["--a-float"],
                help="> Type: float, Default: 3.14",
                required=False,
                default=3.14,
            ),
        ),
        # bool
        (
            "a_bool",
            bool,
            MISSING,
            ArgparseArg(
                option_strings=["--a-bool"],
                help="> Type: bool, Default: None",
                nargs=0,
                const=True,
                required=False,
                default=None,
            ),
        ),
        # bool (True)
        (
            "a_bool",
            bool,
            True,
            ArgparseArg(
                option_strings=["--a-bool"],
                help="> Type: bool, Default: True",
                nargs=0,
                const=True,
                required=False,
                default=True,
            ),
        ),
        # list
        (
            "a_list",
            List[str],
            MISSING,
            ArgparseArg(
                option_strings=["--a-list"],
                help="> Type: List[str], Required",
                nargs="+",
                required=True,
            ),
        ),
        (
            "a_list",
            List[str],
            list,
            ArgparseArg(
                option_strings=["--a-list"],
                help="> Type: List[str], Default: []",
                nargs="*",
                required=False,
                default=[],
            ),
        ),
        # tuple
        (
            "a_tuple",
            Tuple[str],
            MISSING,
            ArgparseArg(
                option_strings=["--a-tuple"],
                help="> Type: Tuple[str], Required",
                nargs="+",
                required=True,
            ),
        ),
        (
            "a_tuple",
            Tuple[str],
            tuple,
            ArgparseArg(
                option_strings=["--a-tuple"],
                help="> Type: Tuple[str], Default: ()",
                nargs="*",
                required=False,
                default=(),
            ),
        ),
        # sequence
        (
            "a_sequence",
            Sequence[str],
            MISSING,
            ArgparseArg(
                option_strings=["--a-sequence"],
                help="> Type: Sequence[str], Required",
                nargs="+",
                required=True,
            ),
        ),
        (
            "a_sequence",
            Sequence[str],
            list,
            ArgparseArg(
                option_strings=["--a-sequence"],
                help="> Type: Sequence[str], Default: []",
                nargs="*",
                required=False,
                default=[],
            ),
        ),
        # set
        (
            "a_set",
            Set[str],
            MISSING,
            ArgparseArg(
                option_strings=["--a-set"],
                help="> Type: Set[str], Required",
                nargs="+",
                required=True,
            ),
        ),
        (
            "a_set",
            Set[str],
            set,
            ArgparseArg(
                option_strings=["--a-set"],
                help="> Type: Set[str], Default: set()",
                nargs="*",
                required=False,
                default=set(),
            ),
        ),
        # dict
        (
            "a_dict",
            Dict[str, str],
            MISSING,
            ArgparseArg(
                option_strings=["--a-dict"],
                help="> Type: Dict[str, str], Required",
                nargs="+",
                required=True,
            ),
        ),
        (
            "a_dict",
            Dict[str, str],
            dict,
            ArgparseArg(
                option_strings=["--a-dict"],
                help="> Type: Dict[str, str], Default: {}",
                nargs="*",
                required=False,
                default={},
            ),
        ),
        # mapping
        (
            "a_mapping",
            Mapping[str, str],
            MISSING,
            ArgparseArg(
                option_strings=["--a-mapping"],
                help="> Type: Mapping[str, str], Required",
                nargs="+",
                required=True,
            ),
        ),
        (
            "a_mapping",
            Mapping[str, str],
            dict,
            ArgparseArg(
                option_strings=["--a-mapping"],
                help="> Type: Mapping[str, str], Default: {}",
                nargs="*",
                required=False,
                default={},
            ),
        ),
        (
            "a_literal",
            Literal["one", "two", "three"],
            "two",
            ArgparseArg(
                option_strings=["--a-literal"],
                help="> Type: Literal['one', 'two', 'three'], Default: 'two'",
                choices=["one", "two", "three"],
                required=False,
                default="two",
            ),
        ),
        (
            "a_literal_int",
            Literal[1, 2, 3],
            2,
            ArgparseArg(
                option_strings=["--a-literal-int"],
                help="> Type: Literal[1, 2, 3], Default: 2",
                choices=[1, 2, 3],
                required=False,
                default=2,
            ),
        ),
    ],
)
def test_add_arguments(
    name: str,
    type_annotation: Type[Any],
    default: Any,
    expected: ArgparseArg,
):
    # 1. ARRANGE
    expected.dest = name

    if not isinstance(type_annotation, _AnnotatedAlias):
        default = yapx.arg(default=default)

    ArgsModel: Type[yapx.types.Dataclass] = make_dataclass(
        "ArgsModel",
        [(name, type_annotation, default)],
    )

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)

    # 3. ASSERT
    # pylint: disable=protected-access
    parser_actions: List[Action] = parser._actions[2:]
    # pylint: disable=no-member
    assert len(parser_actions) == len(ArgsModel.__dataclass_fields__)

    for action in parser_actions:
        assert action.dest

        action_dict: Dict[str, Any] = vars(action)

        expected_dict: Dict[str, Any] = expected.asdict()
        for custom_key in "action", "pos", "group", "exclusive", "type":
            del expected_dict[custom_key]

        for k, v in expected_dict.items():
            assert action_dict[k] == v, f"value of {k} is not equal for '{action.dest}'"


def test_add_arguments_twice():
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        value: str

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(ArgsModel)

    # 3. ASSERT
    with pytest.raises(ParserClosedError):
        parser.add_arguments(ArgsModel)


def test_add_arguments_conflict():
    # 1. ARRANGE
    @dataclass
    class ArgsModel:
        conflict: str = yapx.arg(flags=["--value"])
        value: Optional[str] = None

    expected: str = "hello world"
    cli_args: List[str] = ["--value", expected]

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    with pytest.raises(ArgumentError):
        parser.add_arguments(ArgsModel)

    result: ArgsModel = parser.parse_args_to_model(cli_args)

    # 3. ASSERT
    assert not result.value
    assert result.conflict == expected


def test_add_arguments_func():
    # 1. ARRANGE
    # pylint: disable=unused-argument
    def func(value: str = yapx.arg(default="hello world")):
        ...

    expected: str = "hello world"

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(func)
    args: Dict[str, Any] = vars(parser.parse_args([]))

    # 3. ASSERT
    assert args
    assert "value" in args
    assert args["value"] == expected


@pytest.mark.skipif(
    sys.version_info.minor < 10,
    reason="test modern annotations in Python 3.10+",
)
def test_add_arguments_modern():
    # pylint: disable=unused-argument,unsupported-binary-operation,unsubscriptable-object
    # 1. ARRANGE
    # noinspection Annotator
    def func(value: None | list[str] = yapx.arg(default=lambda: ["hello world"])):
        ...

    expected: list[str] = ["hello world"]

    # 2. ACT
    parser: yapx.ArgumentParser = yapx.ArgumentParser()
    parser.add_arguments(func)
    args: dict[str, str] = vars(parser.parse_args([]))

    # 3. ASSERT
    assert args
    assert "value" in args
    assert args["value"] == expected



import os
import re
import sys
from enum import Enum, auto
from ipaddress import IPv4Address
from pathlib import Path
from typing import Any, Dict, List, Optional, Pattern, Sequence, Set, Tuple
from unittest import mock

import pytest
from _pytest.capture import CaptureFixture, CaptureResult

import yapx
import yapx.argument_parser
from yapx.exceptions import MutuallyExclusiveArgumentError

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


def example_setup(text: str = "world") -> str:
    print("hello " + text)


def example_setup_generator(text: str = "world") -> str:
    print("hello " + text)
    yield
    print("hallo " + text)
    yield
    yield
    yield


def example_empty_subcmd() -> str:
    ...


def example_subcmd(name: str, upper: Optional[bool]) -> str:
    msg: str = "howdy " + name
    if upper:
        msg = msg.upper()
    print(msg)


def test_run_noargs(capsys: CaptureFixture):
    # 1. ARRANGE
    expected: str = "hello world"
    cli_args: List[str] = []

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(example_setup)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    assert captured.out.strip() == expected


def test_run_default(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = ["--text", text]
    expected: List[str] = ["hello " + text]
    not_expected: List[str] = ["howdy"]

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(example_setup, example_subcmd)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_command(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = ["example-subcmd", "--name", text, "--upper"]
    expected: List[str] = [f"howdy {text}".upper()]
    not_expected: List[str] = ["hello"]

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run_command(example_subcmd)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_subcmd(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = ["example-subcmd", "--name", text, "--upper"]
    expected: List[str] = [f"howdy {text}".upper()]
    not_expected: List[str] = ["hello"]

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(None, example_subcmd)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_both(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = [
        "--text",
        text,
        "example-subcmd",
        "--name",
        text,
        "--upper",
    ]
    expected: List[str] = [f"hello {text}", f"howdy {text}".upper()]
    not_expected: List[str] = []

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(example_setup, example_subcmd)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_generator(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = [
        "--text",
        text,
        "example-subcmd",
        "--name",
        text,
        "--upper",
    ]
    expected: List[str] = [f"hello {text}", f"howdy {text}".upper(), f"hallo {text}"]
    not_expected: List[str] = []

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(example_setup_generator, example_subcmd)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_args(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = [
        "--text",
        text,
        "example-subcmd",
        "--name",
        text,
        "--upper",
    ]
    expected: List[str] = [f"hello {text}", f"howdy {text}".upper()]
    not_expected: List[str] = []

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(example_setup, example_subcmd, _args=cli_args)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_kwargs_alias(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = [
        "--text",
        text,
        "command-alias",
        "--name",
        text,
        "--upper",
    ]
    expected: List[str] = [f"hello {text}", f"howdy {text}".upper()]
    not_expected: List[str] = []

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(
            example_setup,
            **{"command-alias": example_subcmd},
        )

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_run_kwargs_alias2(capsys: CaptureFixture):
    # 1. ARRANGE
    text: str = "donald"
    cli_args: List[str] = [
        "--text",
        text,
        "command-alias",
        "--name",
        text,
        "--upper",
    ]
    expected: List[str] = [f"hello {text}", f"howdy {text}".upper()]
    not_expected: List[str] = []

    # 2. ACT
    with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
        yapx.run(example_setup, command_alias=example_subcmd)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_ipv4address(use_pydantic: bool):
    env_var_name: str = "FUNKY_ARG"

    env_values: List[str] = ["127.0.0.1", "192.168.0.1", "9.9.9.9"]
    os.environ[env_var_name] = " LIST[ " + "   ".join(env_values) + " ] "

    expected: List[IPv4Address] = [IPv4Address(ip) for ip in env_values]

    def _func(
        ip_addrs: List[IPv4Address] = yapx.arg(env=env_var_name),
    ) -> List[int]:
        return ip_addrs

    del os.environ[env_var_name]

    # 1. ARRANGE
    cli_args: List[str] = []

    # 2. ACT
    with mock.patch.object(
        yapx.argument_parser,
        "is_pydantic_available",
        mock.Mock(return_value=use_pydantic),
    ), mock.patch.object(
        yapx.argument_parser.sys,
        "argv",
        [""] + cli_args,
    ):
        result: List[Any] = yapx.run(_func)

        # 3. ASSERT
        assert result == expected


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_enum(use_pydantic: bool):
    # 1. ARRANGE
    class MyEnum(Enum):
        one = auto()
        two = auto()
        three = auto()

    def _func(
        value: Optional[MyEnum],
        value_seq: Optional[Sequence[MyEnum]],
        value_default: Optional[MyEnum] = MyEnum.two,
        value_seq_default: Optional[Sequence[MyEnum]] = yapx.arg(
            default=lambda: [MyEnum.three],
        ),
        value_int_default: Optional[Sequence[int]] = yapx.arg(
            default=lambda: [3, 2, 1],
        ),
    ) -> Tuple[Any, ...]:
        return value, value_seq, value_default, value_seq_default, value_int_default

    cli_args = ["--value", "one", "--value-seq", "three", "two", "one"]

    expected: Tuple[Any, ...] = (
        MyEnum.one,
        [MyEnum.three, MyEnum.two, MyEnum.one],
        MyEnum.two,
        [MyEnum.three],
        [3, 2, 1],
    )

    # 2. ACT
    with mock.patch.object(
        yapx.argument_parser,
        "is_pydantic_available",
        mock.Mock(return_value=use_pydantic),
    ), mock.patch.object(
        yapx.argument_parser.sys,
        "argv",
        [""] + cli_args,
    ):
        result: List[Any] = yapx.run(_func)

    # 3. ASSERT
    assert result

    for i in range(len(expected)):
        assert result[i] == expected[i]


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_path(use_pydantic: bool):
    def _func(
        value: Optional[Path],
        value_seq: Optional[Sequence[Path]],
        value_default: Optional[Path] = Path.cwd(),
        value_seq_default: Optional[Sequence[Path]] = yapx.arg(
            default=lambda: [Path.cwd()],
        ),
    ) -> Tuple[Any, ...]:
        return value, value_seq, value_default, value_seq_default

    cli_args = [
        "--value",
        ".",
        "--value-seq",
        str(Path.cwd()),
        str(Path.cwd().parent),
        str(Path.home()),
    ]

    expected: Tuple[Any, ...] = (
        Path("."),
        [Path.cwd(), Path.cwd().parent, Path.home()],
        Path.cwd(),
        [Path.cwd()],
    )

    # 2. ACT
    with mock.patch.object(
        yapx.argument_parser,
        "is_pydantic_available",
        mock.Mock(return_value=use_pydantic),
    ), mock.patch.object(
        yapx.argument_parser.sys,
        "argv",
        [""] + cli_args,
    ):
        result: List[Any] = yapx.run(_func)

    # 3. ASSERT
    assert result

    for i in range(len(expected)):
        assert result[i] == expected[i]


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_patterns(use_pydantic: bool):
    env_var_name: str = "FUNKY_ARG"

    env_values: List[str] = ["abc", "def", ".*"]
    os.environ[env_var_name] = (
        " LiSt[ " + "   ".join(str(i) for i in env_values) + " ] "
    )

    expected: List[Pattern] = [re.compile(x) for x in env_values]

    def _func(ip_addrs: List[Pattern] = yapx.arg(env=env_var_name)) -> List[int]:
        return ip_addrs

    del os.environ[env_var_name]

    # 1. ARRANGE
    cli_args: List[str] = []

    # 2. ACT
    with mock.patch.object(
        yapx.argument_parser,
        "is_pydantic_available",
        mock.Mock(return_value=use_pydantic),
    ), mock.patch.object(
        yapx.argument_parser.sys,
        "argv",
        [""] + cli_args,
    ):
        if not use_pydantic:
            with pytest.raises(yapx.exceptions.UnsupportedTypeError):
                yapx.run(_func)
        else:
            result: List[Any] = yapx.run(_func)

            # 3. ASSERT
            assert result == expected


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_pattern(use_pydantic: bool):
    env_var_name: str = "FUNKY_ARG"

    env_value: str = r".+"
    os.environ[env_var_name] = env_value

    expected: Pattern = re.compile(env_value)

    def _func(ip_addr: Pattern = yapx.arg(env=env_var_name)) -> Pattern:
        return ip_addr

    del os.environ[env_var_name]

    # 1. ARRANGE
    cli_args: List[str] = []

    # 2. ACT
    with mock.patch.object(
        yapx.argument_parser,
        "is_pydantic_available",
        mock.Mock(return_value=use_pydantic),
    ), mock.patch.object(
        yapx.argument_parser.sys,
        "argv",
        [""] + cli_args,
    ):
        if not use_pydantic:
            with pytest.raises(yapx.exceptions.UnsupportedTypeError):
                yapx.run(_func)
        else:
            result: List[Any] = yapx.run(_func)

            # 3. ASSERT
            assert result == expected


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_bools(use_pydantic: bool):
    env_var_name: str = "FUNKY_ARG"

    env_values: List[str] = ["0", "1", "true", "t", "false", "f", "yes", "y", "no", "n"]
    os.environ[env_var_name] = (
        " list[ " + "   ".join(str(i) for i in env_values) + " ] "
    )

    expected: List[bool] = [
        x.lower() in ("1", "true", "t", "yes", "y") for x in env_values
    ]

    def _func(ip_addrs: List[bool] = yapx.arg(env=env_var_name)) -> List[int]:
        return ip_addrs

    del os.environ[env_var_name]

    # 1. ARRANGE
    cli_args: List[str] = []

    # 2. ACT
    try:
        if not use_pydantic:
            mock.patch.object(
                yapx.argument_parser.create_pydantic_model_from_dataclass,
                attribute="__module__",
                new_callable=mock.PropertyMock(return_value="yapx.argument_parser"),
            ).start()

        with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
            result: List[Any] = yapx.run(_func)
    finally:
        mock.patch.stopall()

    # 3. ASSERT
    assert result
    assert isinstance(result, list)

    assert result == expected


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_pos_list(use_pydantic: bool):
    def _func(
        this: Optional[List[str]] = yapx.arg(None, pos=True, exclusive=True),
        that: Optional[List[str]] = yapx.arg(None, exclusive=True),
    ) -> List[int]:
        return this, that

    # 1. ARRANGE
    cli_args: List[str] = ["--that", "world"]

    # 2. ACT
    try:
        if not use_pydantic:
            mock.patch.object(
                yapx.argument_parser.create_pydantic_model_from_dataclass,
                attribute="__module__",
                new_callable=mock.PropertyMock(return_value="yapx.argument_parser"),
            ).start()

        with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
            result: List[Any] = yapx.run(_func)
    finally:
        mock.patch.stopall()

    # 3. ASSERT
    assert result


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_exclusive(use_pydantic: bool):
    def _func(
        this: Optional[List[str]] = yapx.arg(None, pos=True, exclusive=True),
        that: Optional[List[str]] = yapx.arg(None, exclusive=True),
        the: Optional[bool] = yapx.arg(None, exclusive=True),
    ) -> List[int]:
        return this, that, the

    # 1. ARRANGE
    cli_args: List[str] = ["hello", "--that", "world"]

    # 2. ACT
    try:
        if not use_pydantic:
            mock.patch.object(
                yapx.argument_parser.create_pydantic_model_from_dataclass,
                attribute="__module__",
                new_callable=mock.PropertyMock(return_value="yapx.argument_parser"),
            ).start()

        with mock.patch.object(
            yapx.argument_parser.sys,
            "argv",
            [""] + cli_args,
        ), pytest.raises(MutuallyExclusiveArgumentError):
            yapx.run(_func)
    finally:
        mock.patch.stopall()


def test_print_shell_completion(capsys: CaptureFixture):
    # 1. ARRANGE
    cli_args: List[str] = ["--print-shell-completion", "zsh"]
    expected: List[str] = ["--text", "example-empty-subcmd", "example-subcmd"]
    not_expected: List[str] = []

    # 2. ACT
    with pytest.raises(SystemExit):
        yapx.run(
            example_setup,
            example_empty_subcmd,
            example_subcmd,
            _args=cli_args,
        )

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()

    assert captured.out
    for e in expected:
        assert e in captured.out
    for ne in not_expected:
        assert ne not in captured.out


def test_extra_args():
    # 1. ARRANGE
    cli_args: List[str] = ["subcmd", "what", "in", "the", "world=this"]
    expected: List[str] = cli_args[1:]

    def _setup(*args, _extra_args: Optional[List[str]] = None, **kwargs) -> str:
        assert args == tuple(_extra_args)
        assert _extra_args[0] in kwargs
        assert kwargs["world"] == "this"
        return _extra_args

    def _subcmd(_relay_value: Any) -> str:
        assert _relay_value == expected
        return _relay_value

    # 2. ACT
    result = yapx.run(
        _setup,
        _subcmd,
        _args=cli_args,
    )

    # 3. ASSERT
    assert result == expected


def test_annotated():
    # 1. ARRANGE
    expected_value: int = 123
    extra_args: List[str] = ["what", "in", "the", "world"]

    def _setup(
        *args,
        value: Annotated[int, yapx.arg(expected_value)],
        value2: int = yapx.arg(expected_value),
        value3: Annotated[Optional[int], yapx.arg(expected_value)] = None,
        value4: Annotated[float, yapx.arg()] = expected_value,
    ) -> str:
        return args, value, value2, value3, value4

    # 2. ACT
    result = yapx.run(
        _setup,
        _args=extra_args,
    )

    # 3. ASSERT
    assert result[0] == tuple(extra_args)

    for x in result[1:]:
        assert x == expected_value


@pytest.mark.parametrize("use_pydantic", [False, True])
def test_run_everything(use_pydantic: bool):
    def _setup():
        return "hello_relay"

    def _func(
        *args: str,
        _extra_args: List[str],
        v1,
        v2: str,
        v3: Annotated[str, yapx.arg("hello_v3")],
        v4: str = "hello_v4",
        v5: str = yapx.arg("hello_v5"),
        #
        v6: Optional[int] = None,
        v7: Optional[Annotated[int, yapx.arg(7)]] = None,
        v8: int = 8,
        v9: int = yapx.arg(9),
        #
        v10: Optional[float] = None,
        v11: Optional[Annotated[float, yapx.arg(3.11)]] = None,
        v12: float = 3.12,
        v13: float = yapx.arg(3.13),
        #
        v14: Optional[bool] = None,
        v15: Optional[Annotated[bool, yapx.arg(False)]] = None,
        v16: Optional[Annotated[bool, yapx.arg(True, action="store_false")]] = None,
        v17: bool = False,
        v18: bool = yapx.arg(False),
        #
        v19: Optional[List[float]] = None,
        v20: Optional[Annotated[List[float], yapx.arg(lambda: [3.2])]] = None,
        v21: List[float] = lambda: [3.21],
        v22: List[float] = yapx.arg(lambda: [3.22]),
        #
        v23: Optional[Sequence[float]] = None,
        v24: Optional[Annotated[Sequence[float], yapx.arg(lambda: [3.2])]] = None,
        v25: Sequence[float] = lambda: [3.21],
        v26: Sequence[float] = yapx.arg(lambda: [3.22]),
        #
        v27: Optional[Tuple[float, ...]] = None,
        v28: Optional[Annotated[Tuple[float, ...], yapx.arg(lambda: (3.2,))]] = None,
        v29: Tuple[float, ...] = lambda: (3.21,),
        v30: Tuple[float, ...] = yapx.arg(lambda: (3.22,)),
        #
        v31: Optional[Set[float]] = None,
        v32: Optional[Annotated[Set[float], yapx.arg(lambda: {3.2})]] = None,
        v33: Set[float] = lambda: {3.21},
        v34: Set[float] = yapx.arg(lambda: {3.22}),
        #
        v35: Optional[Dict[str, float]] = None,
        v36: Optional[
            Annotated[Dict[str, float], yapx.arg(lambda: {"hello": 3.2})]
        ] = None,
        v37: Dict[str, float] = lambda: {"hello": 3.21},
        v38: Dict[str, float] = yapx.arg(lambda: {"hello": 3.22}),
        #
        _relay_value: Any = None,
        **kwargs: Optional[str],
    ) -> None:
        assert args
        assert args == tuple(_extra_args)
        assert args[0] in kwargs
        assert _relay_value == "hello_relay"

        assert v1 == "hello_v1"
        assert v2 == "hello_v2"
        assert v3 == "hello_v3"
        assert v4 == "hello_v4"
        assert v5 == "hello_v5"

        assert v6 == 6
        assert v7 == 7
        assert v8 == 8
        assert v9 == 9

        assert v10 == 3.1
        assert v11 == 3.11
        assert v12 == 3.12
        assert v13 == 3.13

        assert v14 is True
        assert v15 is False
        assert v16 is False
        assert v17 is False
        assert v18 is False

        # list
        assert v19 == [3.19, 3.19, 3.192]
        assert v20 == [3.2]
        assert v21 == [3.21]
        assert v22 == [3.22]

        # sequence
        assert v23 == [3.19, 3.19, 3.192]
        assert v24 == [3.2]
        assert v25 == [3.21]
        assert v26 == [3.22]

        # tuple
        assert v27 == (3.19, 3.19, 3.192)
        assert v28 == (3.2,)
        assert v29 == (3.21,)
        assert v30 == (3.22,)

        # set
        assert v31 == {3.19, 3.192}
        assert v32 == {3.2}
        assert v33 == {3.21}
        assert v34 == {3.22}

        # dict
        assert v35 == {"hello": 3.19, "world": 3.192}
        assert v36 == {"hello": 3.2}
        assert v37 == {"hello": 3.21}
        assert v38 == {"hello": 3.22}

    cli_args: List[str] = [
        "func",
        "--v1",
        "hello_v1",
        "--v2",
        "hello_v2",
        "--v6",
        "6",
        "--v10",
        "3.10",
        "--v14",
        "--v16",
        "--v19",
        "3.19",
        "3.19",
        "3.192",
        "--v23",
        "3.19,",
        "3.19,",
        "3.192",
        "--v27",
        "3.19,",
        "3.19,",
        "3.192",
        "--v31",
        "3.19,",
        "3.19,",
        "3.192",
        "--v35",
        "hello=3.19,",
        "world=3.192",
        "--",
        "purposefully_extra",
        "purposefully_hello=world",
    ]

    # 2. ACT
    try:
        if not use_pydantic:
            mock.patch.object(
                yapx.argument_parser.create_pydantic_model_from_dataclass,
                attribute="__module__",
                new_callable=mock.PropertyMock(return_value="yapx.argument_parser"),
            ).start()

        with mock.patch.object(yapx.argument_parser.sys, "argv", [""] + cli_args):
            yapx.run(_setup, _func)
    finally:
        mock.patch.stopall()



from typing import Any, Optional, Type


class ArgumentConflictError(Exception):
    ...


class ParserClosedError(Exception):
    ...


class NoArgsModelError(Exception):
    ...


class MutuallyExclusiveArgumentError(Exception):
    ...


class MutuallyExclusiveRequiredError(Exception):
    ...


class UnsupportedTypeError(TypeError):
    ...


def raise_unsupported_type_error(
    fld_type: Type[Any],
    from_exception: Optional[Exception] = None,
):
    fld_type_name: str = (
        fld_type.__name__ if hasattr(fld_type, "__name__") else str(fld_type)
    )

    msg: str = (
        f"Unsupported type: {fld_type_name}\nInstall 'pydantic' to support more types."
    )

    if from_exception:
        raise UnsupportedTypeError(msg) from from_exception
    raise UnsupportedTypeError(msg)



import os
from dataclasses import MISSING, Field, fields, is_dataclass
from typing import Any, Dict, Optional, Tuple, Type

import pytest

import yapx
from yapx.arg import (
    ARGPARSE_ARG_METADATA_KEY,
    ArgparseArg,
    convert_to_command_string,
    convert_to_flag_string,
    convert_to_short_flag_string,
)


def test_convert_to_command_string() -> None:
    assert convert_to_command_string(" this_THAT ") == "this-that"
    assert convert_to_command_string(" _this_that ") == "that"
    assert convert_to_command_string(" x_that ") == "that"


def test_convert_to_flag_string() -> None:
    assert convert_to_flag_string("test") == "--test"
    assert convert_to_flag_string("  -TeST  ") == "--test"
    assert convert_to_flag_string("test_this") == "--test-this"


def test_convert_to_short_flag_string() -> None:
    assert convert_to_short_flag_string("test") == "-t"
    assert convert_to_short_flag_string("  --TeST  ") == "-t"
    assert convert_to_short_flag_string("test_this") == "-t"


def test_argument_parser_arg_defaults():
    # 1. ARRANGE
    expected: Dict[str, Any] = {
        "dest": "test_this",
        "option_strings": ["--test-this"],
        "type": None,
        "required": True,
        "group": None,
        "exclusive": False,
        "pos": False,
        "nargs": None,
        "const": None,
        "default": None,
        "choices": None,
        "help": None,
        "metavar": None,
        "action": None,
    }

    # 2. ACT
    arg: ArgparseArg = ArgparseArg(dest=expected["dest"])

    # 3. ASSERT
    assert arg
    assert is_dataclass(arg)
    assert arg.asdict() == expected


# pylint: disable=dangerous-default-value
def test_build_dataclass_from_func_raises_type_error():
    # 1. ARRANGE
    # pylint: disable=unused-argument,too-many-arguments
    def func1(value=lambda: []):
        pass

    def func2(value=lambda: {}):
        pass

    def func3(value=lambda: ()):
        pass

    def func4(value=b""):
        pass

    # 2,3. ACT & ASSERT
    for func in func1, func2, func3, func4:
        with pytest.raises(TypeError):
            yapx.utils.make_dataclass_from_func(func)


def _assert_model_field_attributes(
    model: Type[yapx.types.Dataclass],
    expected_field_props: Dict[str, Dict[str, Any]],
    expected_metadata_field_props: Dict[str, Dict[str, Any]],
):
    model_fields: Tuple[Field, ...] = fields(model)

    for f in model_fields:
        assert f.name in expected_field_props
        for exp_field_prop, exp_field_prop_value in expected_field_props[
            f.name
        ].items():
            assert getattr(f, exp_field_prop) == exp_field_prop_value

        assert ARGPARSE_ARG_METADATA_KEY in f.metadata
        f_metadata = f.metadata[ARGPARSE_ARG_METADATA_KEY]
        assert isinstance(f_metadata, ArgparseArg)

        assert f.name in expected_metadata_field_props
        assert f_metadata.asdict() == expected_metadata_field_props[f.name]


def test_build_dataclass_from_func_no_anno():
    # 1. ARRANGE
    expected_field_props: Dict[str, Dict[str, Any]] = {
        "value": {
            "name": "value",
            "type": str,
            "default": MISSING,
        },
    }
    expected_metadata_field_props: Dict[str, Dict[str, Any]] = {
        "value": ArgparseArg().asdict(),
    }

    # 2. ACT
    model: Type[yapx.types.Dataclass] = yapx.utils.make_dataclass_from_func(
        lambda value: ...,
    )

    # 3. ASSERT
    _assert_model_field_attributes(
        model,
        expected_field_props,
        expected_metadata_field_props,
    )


def test_build_dataclass_from_func_none_anno():
    # 1. ARRANGE
    # pylint: disable=unused-argument,too-many-arguments
    def func(value=None):
        """testing docstring to argparse description"""
        ...

    expected_field_props: Dict[str, Dict[str, Any]] = {
        "value": {
            "name": "value",
            "type": Optional[str],
            "default": None,
        },
    }
    expected_metadata_field_props: Dict[str, Dict[str, Any]] = {
        "value": ArgparseArg(required=False).asdict(),
    }

    # 2. ACT
    model: Type[yapx.types.Dataclass] = yapx.utils.make_dataclass_from_func(func)

    # 3. ASSERT
    _assert_model_field_attributes(
        model,
        expected_field_props,
        expected_metadata_field_props,
    )

    assert model.__doc__ == func.__doc__


def test_build_dataclass_from_func_default():
    # 1. ARRANGE
    # pylint: disable=unused-argument,too-many-arguments
    def func(value=3.14):
        pass

    expected_field_props: Dict[str, Dict[str, Any]] = {
        "value": {
            "name": "value",
            "type": float,
            "default": 3.14,
        },
    }
    expected_metadata_field_props: Dict[str, Dict[str, Any]] = {
        "value": ArgparseArg(required=False, default=3.14).asdict(),
    }

    # 2. ACT
    model: Type[yapx.types.Dataclass] = yapx.utils.make_dataclass_from_func(func)

    # 3. ASSERT
    _assert_model_field_attributes(
        model,
        expected_field_props,
        expected_metadata_field_props,
    )


def test_build_dataclass_from_func_argfield_default():
    # 1. ARRANGE
    # pylint: disable=unused-argument,too-many-arguments
    def func(value=yapx.arg(default=3.14)):
        pass

    expected_field_props: Dict[str, Dict[str, Any]] = {
        "value": {
            "name": "value",
            "type": float,
            "default": 3.14,
        },
    }
    expected_metadata_field_props: Dict[str, Dict[str, Any]] = {
        "value": ArgparseArg(required=False, default=3.14).asdict(),
    }

    # 2. ACT
    model: Type[yapx.types.Dataclass] = yapx.utils.make_dataclass_from_func(func)

    # 3. ASSERT
    _assert_model_field_attributes(
        model,
        expected_field_props,
        expected_metadata_field_props,
    )


def test_build_dataclass_from_func_argfield_envvar():
    # 1. ARRANGE
    env_var_name: str = "FUNKY_ARG"
    os.environ[env_var_name] = "3.1415"

    # pylint: disable=unused-argument,too-many-arguments
    def func(value=yapx.arg(env=env_var_name)):
        pass

    del os.environ[env_var_name]

    expected_field_props: Dict[str, Dict[str, Any]] = {
        "value": {
            "name": "value",
            "type": str,
            "default": "3.1415",
        },
    }
    expected_metadata_field_props: Dict[str, Dict[str, Any]] = {
        "value": ArgparseArg(required=False, default="3.1415").asdict(),
    }

    # 2. ACT
    model: Type[yapx.types.Dataclass] = yapx.utils.make_dataclass_from_func(func)

    # 3. ASSERT
    _assert_model_field_attributes(
        model,
        expected_field_props,
        expected_metadata_field_props,
    )


def test_build_dataclass_from_func_argfield_envvar_file(clean_dir):
    # 1. ARRANGE
    env_var_name: str = "FUNKY_ARG"
    env_var_name_file: str = env_var_name + "_FILE"
    tmp_file: str = "test.tmp"
    os.environ[env_var_name_file] = tmp_file
    with open(tmp_file, "w+", encoding="utf8") as f:
        f.write("3.1415")

    # pylint: disable=unused-argument,too-many-arguments
    def func(value=yapx.arg(env=env_var_name)):
        pass

    del os.environ[env_var_name_file]
    os.remove(tmp_file)

    expected_field_props: Dict[str, Dict[str, Any]] = {
        "value": {
            "name": "value",
            "type": str,
            "default": "3.1415",
        },
    }
    expected_metadata_field_props: Dict[str, Dict[str, Any]] = {
        "value": ArgparseArg(required=False, default="3.1415").asdict(),
    }

    # 2. ACT
    model: Type[yapx.types.Dataclass] = yapx.utils.make_dataclass_from_func(func)

    # 3. ASSERT
    _assert_model_field_attributes(
        model,
        expected_field_props,
        expected_metadata_field_props,
    )
